/****************************************************************/
/*								*/
/* coor.c:	coordinate transformations and data		*/
/*		manipulation					*/
/*								*/
/*		Design: Walter Benzing 1994			*/
/*								*/
/*		NO RIGHTS RESERVED				*/
/****************************************************************/

/* standard includes */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

/* my includes */

#include "xdim.h"

/*
 * allocate a new buffer in act data
 */

int CO_NewData(int NewWidth, int NewHeight, int DataType, XDimInfo *XDim)
   {
    size_t Size;
    ViewportInfo *view;
    void *buffer;

    if(XDim->port >=0)
      {
       view = &((XDim->Views)[XDim->port]);
       view->changed |= DATA_AND_CHANGE;
       switch (DataType)
          {
           case MY_UCHAR:
           Size = NewWidth * NewHeight;
           break;
           case MY_TRIANGLE:
           /*
            * NewWidth = number of elements (=triangles)
            * NewHeight = number of points per element (triangle => 3)
            * Additional space for the color info is neded (=>+COLOR_3D)
            * 3 values per point (x,y,z)
            */
	   Size = NewWidth * (3*NewHeight+COLOR_3D) * sizeof(double);
	   break;
           default:
	   Size = NewWidth * NewHeight * sizeof(double);
	   break;
	  }
	if(Size != 0)
	  {
	   buffer = malloc(Size);
	   if(!buffer)
	     {
	      sprintf(XDim->error, XDERR_NO_MEM_D,
		      (int)Size);
	      return(ERROR);
	     }
	  }
	else
	  {
	   NewWidth = 0;
	   NewHeight = 0;
	  }
       if(view->charData)
         {
          free(view->charData);
          view->charData = NULL;
         }
       if(view->actData)
         {
          free(view->actData);
          view->actData = NULL;
         }
       switch (DataType)
          {
           case MY_UCHAR:
           view->charData = (unsigned char *) buffer;;
           break;
           default:
	   view->actData = (double *) buffer;;
	   break;
	  }
       view->Width = NewWidth;
       view->Height = NewHeight;
       view->dataType = DataType;
      }
    else
      {
       strcpy(XDim->error, XDERR_NO_VIEW);
       return(ERROR);
      }
    return(OK);
   }

/*
 * free data
 */

void CO_FreeData(ViewportInfo *view)
   {
    if(view->charData)
      {
       free(view->charData);
       view->charData = NULL;
      }
    if(view->actData)
      {
       free(view->actData);
       view->actData = NULL;
      }
    if(view->Data)
      {
       free(view->Data);
       view->Data = NULL;
      }
   }

/*
 * find Max, Min
 */

void CO_MinMax(double *max, double *min, int Size, double *Data)
   {
    int i;
    double Max, Min, *dp, val;

    dp = Data;
    Max = *dp;
    Min = *dp++;
    for(i=1; i<Size; i++)
       {
	val = *dp++;
	if(val<Min)
	  Min = val;
	else if(val>Max)
	  Max = val;
       }
    *max = Max;
    *min = Min;
   }

void CO_MinMaxChar(double *max, double *min, int Size, unsigned char *CData)
   {
    int i;
    unsigned char Max, Min, *cp, val;

    cp = CData;
    Max = *cp;
    Min = *cp++;
    for(i=1; i<Size; i++)
       {
	val = *cp++;
	if(val<Min)
	  Min = val;
	else if(val>Max)
	  Max = val;
       }
    *max = (double)Max;
    *min = (double)Min;
   }

/*
 * find Max, Min for Triangle data
 * num = number of points per element
 * Size = number of elements
 */

void CO_MinMaxTri(double *max, double *min, int which, int num, int Size,
	double *Data)
   {
    int i, j, delta;
    double Max, Min, *dp, val;

    /*
     * which => x=0, y= 1, z =2
     */
    dp = Data+which;
    /*
     * x,y,z => 3
     */
    delta = 3;
    Max = *dp;
    Min = *dp;
    for(i=0; i<Size; i++)
       {
        /* process all points */
        for(j=0; j< num; j++)
           {
	    val = *dp;
	    dp += delta;
	    if(val<Min)
	      Min = val;
	    else if(val>Max)
	      Max = val;
	   }
	/* skip color info */
	dp+=COLOR_3D;
       }
    *max = Max;
    *min = Min;
   }

/*
 * find Max, Min Color for 3D data
 * num = number of points per element
 * Size = number of elements
 */

void CO_MinMaxCol3D(double *max, double *min, int num, int Size,
	double *Data)
   {
    int i, j, delta;
    double Max, Min, *dp, val;

    /*
     * x=0, y= 1, z =2
     */
    delta = 3*num;
    dp = Data+delta;
    Max = *dp;
    Min = *dp;
    for(i=0; i<Size; i++)
       {
        /* process all color info */
        for(j=0; j< COLOR_3D; j++)
           {
	    val = *dp++;
	    if(val<Min)
	      Min = val;
	    else if(val>Max)
	      Max = val;
	   }
	/* skip 3d data */
	dp+=delta;
       }
    *max = Max;
    *min = Min;
   }

/*
 * calculate the base for the axis
 */

double CO_Base(double Max)
   {
    double base;
    
    Max = fabs(Max);
    if(Max > 0.0)
      {
       base = log10(fabs(Max));
       base = floor(base/3.0+0.2)*3.0;
       base = exp(log(10.0)*base);
      }
    else
      base = 1.0;
    return(base);
   }

/*
 * Type cast to double
 */

int CO_Cast2Double(ViewportInfo *view, XDimInfo *XDim)
   {
    int i, Size;
    unsigned char *cp, *pp;
    double *dp;

    switch (view->dataType)
     {
      case MY_UCHAR:
      if(!view->charData)
	{
	 sprintf(XDim->error, XDERR_NO_DATA2DOUBLE_S,
	    TypeNames[MY_UCHAR]);
	 return(ERROR);
	}

      /* CAUTION old data is removed with CO_NewData */
      
      pp = view->charData;
      view->charData = NULL;
      if(CO_NewData(view->Width, view->Height, MY_DOUBLE, XDim))
	 return(ERROR);
      cp = pp;
      dp = view->actData;
      Size = view->Width * view->Height;
      for(i=0; i<Size; i++)
	 {
	  *dp++ = (double)*cp++;
	 }
      free(pp);
      view->charData = NULL;
      break;
      case MY_TRIANGLE:
        {
	 sprintf(XDim->error, XDERR_CAST_TO_DOUBLE_S, TypeNames[MY_TRIANGLE]);
	 return(ERROR);
	}
      default:
      break;
     }
    view->dataType = MY_DOUBLE;
    return(OK);
   }

/*
 * Type cast to unsigned char
 */

int CO_Cast2UChar(ViewportInfo *view, XDimInfo *XDim)
   {
    int i, Size;
    unsigned char *cp;
    double *dp, *pp;

    switch (view->dataType)
     {
      case MY_UCHAR:
      break;
      case MY_TRIANGLE:
        {
	 sprintf(XDim->error, XDERR_CAST_TO_UCHAR_S,
		TypeNames[MY_TRIANGLE]);
	 return(ERROR);
	}
      break;
      default:
      if(!view->actData)
	{
	 sprintf(XDim->error, XDERR_NO_DATA2UCHAR_S,
	    TypeNames[MY_DOUBLE]);
	 return(ERROR);
	}

      /* CAUTION old data is removed with CO_NewData */
      
      pp = view->actData;
      view->actData = NULL;
      if(CO_NewData(view->Width, view->Height, MY_UCHAR, XDim))
	 return(ERROR);
      cp = view->charData;
      dp = pp;
      Size = view->Width * view->Height;
      for(i=0; i<Size; i++)
	 {
	  *cp++ = (unsigned char)*dp++;
	 }
      free(pp);
      view->actData = NULL;
      break;
     }
    view->dataType = MY_UCHAR;
    return(OK);
   }

/*
 * CO_IntersectX removes parts of triangles that lay outside
 * MaxX. It takes information from ip and puts one to
 * ip and one to op. The number of inserted elements is returned
 */

int CO_IntersectX(double *ip, double **to, double MaxX, int *copyAll,
	int *num, int isMax)
   {
    double x0,y0,z0,x1,y1,z1,x2,y2,z2, ys1, zs1, ys2, zs2;
    double c0, c1, c2, cs1, cs2, fact, *op = *to;
    int np,del = 1, inside=False, flag = *copyAll;

    np = CO_NONE;
    if(isMax)
      {
       if((x0 = ip[0]) > MaxX)
	 np = CO_FIRST;
       if((x1 = ip[3]) > MaxX)
	 np |= CO_SECOND;
       if((x2 = ip[6]) > MaxX)
	 np |= CO_THIRD;
      }
    else
      {
       if((x0 = ip[0]) < MaxX)
	 np = CO_FIRST;
       if((x1 = ip[3]) < MaxX)
	 np |= CO_SECOND;
       if((x2 = ip[6]) < MaxX)
	 np |= CO_THIRD;
      }
    switch(np)
      {
       case CO_NONE:
         /* no intersection */
         if(flag)
           {
            memcpy(op, ip, 10*sizeof(double));
            *to += 10;
           }
         return(flag);
       /*
        * Triangle intersects
        * reorder the values so that 0->1 and 0->2 intersect
        */
       case (CO_SECOND | CO_THIRD):
         inside = True;
         del = 0;
       case CO_FIRST:
	 y0 = *(ip+1);
	 z0 = *(ip+2);
	 y1 = *(ip+4);
	 z1 = *(ip+5);
	 y2 = *(ip+7);
	 z2 = *(ip+8);
	 /* color */
	 c0 = *(ip+9);
	 c1 = *(ip+10);
	 c2 = *(ip+11);
	 break;
       case (CO_FIRST | CO_THIRD):
         inside = True;
         del = 0;
       case CO_SECOND:
         y0 = x0;
         x0 = x1;
         x1 = y0;
	 y1 = *(ip+1);
	 z1 = *(ip+2);
	 y0 = *(ip+4);
	 z0 = *(ip+5);
	 y2 = *(ip+7);
	 z2 = *(ip+8);
	 c1 = *(ip+9);
	 c0 = *(ip+10);
	 c2 = *(ip+11);
	 break;
       case (CO_FIRST | CO_SECOND):
         inside = True;
         del = 0;
       case CO_THIRD:
         y0 = x0;
         x0 = x2;
         x2 = y0;
	 y2 = *(ip+1);
	 z2 = *(ip+2);
	 y1 = *(ip+4);
	 z1 = *(ip+5);
	 y0 = *(ip+7);
	 z0 = *(ip+8);
	 c2 = *(ip+9);
	 c1 = *(ip+10);
	 c0 = *(ip+11);
	 break;
       case CO_ALL:
         /* remove triangle */
         (*num)--;
         if(flag)
           return(flag);
         else
           /* first change */
           return(CO_REMOVED);
      }
    /* Calculate Intersection */
    if(x1!=x0)
      {
       fact = (MaxX-x0)/(x1-x0);
       ys1 = y0+ fact * (y1-y0);
       zs1 = z0+ fact * (z1-z0);
       cs1 = c0+ fact * (c1-c0);
      }
    else
      {
       ys1 = y0;
       zs1 = z0;
       cs1 = c0;
      }
    if(x2!=x0)
      {
       fact = (MaxX-x0)/(x2-x0);
       ys2 = y0+ fact * (y2-y0);
       zs2 = z0+ fact * (z2-z0);
       cs2 = c0+ fact * (c2-c0);
      }
    else
      {
       ys2 = y0;
       zs2 = z0;
       cs2 = c0;
      }
    /* Insert Elements */
    if(inside)
      {
       *op++ = x0;
       *op++ = y0;
       *op++ = z0;
       *op++ = MaxX;
       *op++ = ys2;
       *op++ = zs2;
       *op++ = MaxX;
       *op++ = ys1;
       *op++ = zs1;
       *op++ = c0;
       *op++ = cs2;
       *op++ = cs1;
      }
    else
      {
       *op++ = x1;
       *op++ = y1;
       *op++ = z1;
       *op++ = x2;
       *op++ = y2;
       *op++ = z2;
       *op++ = MaxX;
       *op++ = ys1;
       *op++ = zs1;
       *op++ = c1;
       *op++ = c2;
       *op++ = cs1;

       *op++ = x2;
       *op++ = y2;
       *op++ = z2;
       *op++ = MaxX;
       *op++ = ys2;
       *op++ = zs2;
       *op++ = MaxX;
       *op++ = ys1;
       *op++ = zs1;
       *op++ = c2;
       *op++ = cs2;
       *op++ = cs1;
      }
    *to = op;
    *num += del;
    if(flag)
      return(flag);
    else
      return(CO_CHANGED);
   }


/*
 * CO_IntersectY removes parts of triangles that lay outside
 * MaxX. It takes information from ip and puts one to
 * ip and one to op. The number of inserted elements is returned
 */

int CO_IntersectY(double *ip, double **to, double MaxY, int *copyAll,
	int *num, int isMax)
   {
    double x0,y0,z0,x1,y1,z1,x2,y2,z2, xs1, zs1, xs2, zs2;
    double c0, c1, c2, cs1, cs2, fact, *op = *to;
    int np,del = 1, inside=False, flag = *copyAll;

    np = CO_NONE;
    if(isMax)
      {
       if((y0 = ip[1]) > MaxY)
	 np = CO_FIRST;
       if((y1 = ip[4]) > MaxY)
	 np |= CO_SECOND;
       if((y2 = ip[7]) > MaxY)
	 np |= CO_THIRD;
      }
    else
      {
       if((y0 = ip[1]) < MaxY)
	 np = CO_FIRST;
       if((y1 = ip[4]) < MaxY)
	 np |= CO_SECOND;
       if((y2 = ip[7]) < MaxY)
	 np |= CO_THIRD;
      }
    switch(np)
      {
       case CO_NONE:
         /* no intersection */
         if(flag)
           {
            memcpy(op, ip, 10*sizeof(double));
            *to += 10;
           }
         return(flag);
       /*
        * Triangle intersects
        * reorder the values so that 0->1 and 0->2 intersect
        */
       case (CO_SECOND | CO_THIRD):
         inside = True;
         del = 0;
       case CO_FIRST:
	 x0 = *ip;
	 z0 = *(ip+2);
	 x1 = *(ip+3);
	 z1 = *(ip+5);
	 x2 = *(ip+6);
	 z2 = *(ip+8);
	 /* color */
	 c0 = *(ip+9);
	 c1 = *(ip+10);
	 c2 = *(ip+11);
	 break;
       case (CO_FIRST | CO_THIRD):
         inside = True;
         del = 0;
       case CO_SECOND:
         x0 = y0;
         y0 = y1;
         y1 = x0;
	 x1 = *ip;
	 z1 = *(ip+2);
	 x0 = *(ip+3);
	 z0 = *(ip+5);
	 x2 = *(ip+6);
	 z2 = *(ip+8);
	 c1 = *(ip+9);
	 c0 = *(ip+10);
	 c2 = *(ip+11);
	 break;
       case (CO_FIRST | CO_SECOND):
         inside = True;
         del = 0;
       case CO_THIRD:
         x0 = y0;
         y0 = y2;
         y2 = x0;
	 x2 = *ip;
	 z2 = *(ip+2);
	 x1 = *(ip+3);
	 z1 = *(ip+5);
	 x0 = *(ip+6);
	 z0 = *(ip+8);
	 c2 = *(ip+9);
	 c1 = *(ip+10);
	 c0 = *(ip+11);
	 break;
       case CO_ALL:
         /* remove triangle */
         (*num)--;
         if(flag)
           return(flag);
         else
           /* first change */
           return(CO_REMOVED);
      }
    /* Calculate Intersection */
    if(y1!=y0)
      {
       fact = (MaxY-y0)/(y1-y0);
       xs1 = x0+ fact * (x1-x0);
       zs1 = z0+ fact * (z1-z0);
       cs1 = c0+ fact * (c1-c0);
      }
    else
      {
       xs1 = x0;
       zs1 = z0;
       cs1 = c0;
      }
    if(y2!=y0)
      {
       fact = (MaxY-y0)/(y2-y0);
       xs2 = x0+ fact * (x2-x0);
       zs2 = z0+ fact * (z2-z0);
       cs2 = c0+ fact * (c2-c0);
      }
    else
      {
       xs2 = x0;
       zs2 = z0;
       cs2 = c0;
      }
    /* Insert Elements */
    if(inside)
      {
       *op++ = x0;
       *op++ = y0;
       *op++ = z0;
       *op++ = xs2;
       *op++ = MaxY;
       *op++ = zs2;
       *op++ = xs1;
       *op++ = MaxY;
       *op++ = zs1;
       *op++ = c0;
       *op++ = cs2;
       *op++ = cs1;
      }
    else
      {
       *op++ = x1;
       *op++ = y1;
       *op++ = z1;
       *op++ = x2;
       *op++ = y2;
       *op++ = z2;
       *op++ = xs1;
       *op++ = MaxY;
       *op++ = zs1;
       *op++ = c1;
       *op++ = c2;
       *op++ = cs1;

       *op++ = x2;
       *op++ = y2;
       *op++ = z2;
       *op++ = xs2;
       *op++ = MaxY;
       *op++ = zs2;
       *op++ = xs1;
       *op++ = MaxY;
       *op++ = zs1;
       *op++ = c2;
       *op++ = cs2;
       *op++ = cs1;
      }
    *to = op;
    *num += del;
    if(flag)
      return(flag);
    else
      return(CO_CHANGED);
   }


/*
 * CO_IntersectZ removes parts of triangles that lay outside
 * MaxZ. It takes information from ip and puts one to
 * ip and one to op. The number of inserted elements is returned
 */

int CO_IntersectZ(double *ip, double **to, double MaxZ, int *copyAll,
	int *num, int isMax)
   {
    double x0,y0,z0,x1,y1,z1,x2,y2,z2, ys1, xs1, ys2, xs2;
    double c0, c1, c2, cs1, cs2, fact, *op = *to;
    int np,del = 1, inside=False, flag = *copyAll;

    np = CO_NONE;
    if(isMax)
      {
       if((z0 = ip[2]) > MaxZ)
	 np = CO_FIRST;
       if((z1 = ip[5]) > MaxZ)
	 np |= CO_SECOND;
       if((z2 = ip[8]) > MaxZ)
	 np |= CO_THIRD;
      }
    else
      {
       if((z0 = ip[2]) < MaxZ)
	 np = CO_FIRST;
       if((z1 = ip[5]) < MaxZ)
	 np |= CO_SECOND;
       if((z2 = ip[8]) < MaxZ)
	 np |= CO_THIRD;
      }
    switch(np)
      {
       case CO_NONE:
         /* no intersection */
         if(flag)
           {
            memcpy(op, ip, 10*sizeof(double));
            *to += 10;
           }
         return(flag);
       /*
        * Triangle intersects
        * reorder the values so that 0->1 and 0->2 intersect
        */
       case (CO_SECOND | CO_THIRD):
         inside = True;
         del = 0;
       case CO_FIRST:
	 x0 = *ip;
	 y0 = *(ip+1);
	 x1 = *(ip+3);
	 y1 = *(ip+4);
	 x2 = *(ip+6);
	 y2 = *(ip+7);
	 /* color */
	 c0 = *(ip+9);
	 c1 = *(ip+10);
	 c2 = *(ip+11);
	 break;
       case (CO_FIRST | CO_THIRD):
         inside = True;
         del = 0;
       case CO_SECOND:
         y0 = z0;
         z0 = z1;
         z1 = y0;
	 x1 = *ip;
	 y1 = *(ip+1);
	 x0 = *(ip+3);
	 y0 = *(ip+4);
	 x2 = *(ip+6);
	 y2 = *(ip+7);
	 c1 = *(ip+9);
	 c0 = *(ip+10);
	 c2 = *(ip+11);
	 break;
       case (CO_FIRST | CO_SECOND):
         inside = True;
         del = 0;
       case CO_THIRD:
         y0 = z0;
         z0 = z2;
         z2 = y0;
	 x2 = *ip;
	 y2 = *(ip+1);
	 x1 = *(ip+3);
	 y1 = *(ip+4);
	 x0 = *(ip+6);
	 y0 = *(ip+7);
	 c2 = *(ip+9);
	 c1 = *(ip+10);
	 c0 = *(ip+11);
	 break;
       case CO_ALL:
         /* remove triangle */
         (*num)--;
         if(flag)
           return(flag);
         else
           /* first change */
           return(CO_REMOVED);
      }
    /* Calculate Intersection */
    if(z1!=z0)
      {
       fact = (MaxZ-z0)/(z1-z0);
       xs1 = x0+ fact * (x1-x0);
       ys1 = y0+ fact * (y1-y0);
       cs1 = c0+ fact * (c1-c0);
      }
    else
      {
       xs1 = x0;
       ys1 = y0;
       cs1 = c0;
      }
    if(z2!=z0)
      {
       fact = (MaxZ-z0)/(z2-z0);
       ys2 = y0+ fact * (y2-y0);
       xs2 = x0+ fact * (x2-x0);
       cs2 = c0+ fact * (c2-c0);
      }
    else
      {
       xs2 = x0;
       ys2 = y0;
       cs2 = c0;
      }
    /* Insert Elements */
    if(inside)
      {
       *op++ = x0;
       *op++ = y0;
       *op++ = z0;
       *op++ = xs2;
       *op++ = ys2;
       *op++ = MaxZ;
       *op++ = xs1;
       *op++ = ys1;
       *op++ = MaxZ;
       *op++ = c0;
       *op++ = cs2;
       *op++ = cs1;
      }
    else
      {
       *op++ = x1;
       *op++ = y1;
       *op++ = z1;
       *op++ = x2;
       *op++ = y2;
       *op++ = z2;
       *op++ = xs1;
       *op++ = ys1;
       *op++ = MaxZ;
       *op++ = c1;
       *op++ = c2;
       *op++ = cs1;

       *op++ = x2;
       *op++ = y2;
       *op++ = z2;
       *op++ = xs2;
       *op++ = ys2;
       *op++ = MaxZ;
       *op++ = xs1;
       *op++ = ys1;
       *op++ = MaxZ;
       *op++ = c2;
       *op++ = cs2;
       *op++ = cs1;
      }
    *to = op;
    *num += del;
    if(flag)
      return(flag);
    else
      return(CO_CHANGED);
   }

/*
 * This programs transforms triangles
 * from double to myCoor format, if Max, Min
 * etc. lie within the data region, the
 * number of elements (=triangles) changes
 * ip : pointer to input data
 * num: number of input triangles
 * poly: number of points per element;
 * type: data type (interpolated or normal)
 */

my3DInfo *CO_Triangle2Int(double *ip, int num, int poly, int type, double MinX,
   double MaxX, double MinY, double MaxY, double MinZ, double MaxZ,
   double MinC, double MaxC)
   {
    int Size, actSize, numElem, copyAll, offs, delta, i, j, k;
    int actElem, oNorm, oLines, oPoint;
    double *elem, *ep, *elem1, *elem2, *dp, factX, factY, factZ, factC;
    double *doubData, *dop;
    myCoor co2, *data, *op;
    static my3DInfo DDI;
    
    DDI.data = NULL;			/* in case of trouble */
    DDI.inter = 0;			/* interpolation not implemented */
    DDI.norm = -1;			/* error condition */
    DDI.lines = 0;
    DDI.point = 0;
    /*
     * x,y,z => 3
     */
    offs = poly*3+COLOR_3D;
    Size = 64*offs;			/* up to 64 triangles may be created */
    elem1 = malloc(Size*sizeof(double));/* by intersection with sides !! */
    data = NULL;
    doubData = NULL;
    if(elem1 != NULL)
      {
       elem2 = malloc(Size*sizeof(double));
       if(elem2 != NULL)
	 {
	  Size = num*offs;
          if(type)
	     data = malloc(Size*sizeof(myCoor));
	  else
	     doubData = malloc(Size*sizeof(double));
	 }
      }
    if(data || doubData)
      {
       oNorm = 0;
       oLines = 0;			/* not implemented */
       oPoint = 0;			/* not implemented */
       actSize = num;
       if(type)
         {
	  op = data;
	  if(MaxX > MinX)
	    factX = (double)(CO_ONE-1)/(MaxX - MinX);
	  else
	    factX = 1.0;
          if(MaxY > MinY)
	    factY = (double)(CO_ONE-1)/(MaxY - MinY);
          else
	    factY = 1.0;
          if(MaxZ > MinZ)
	    factZ = (double)(CO_ONE-1)/(MaxZ - MinZ);
          else
	    factZ = 1.0;
          if(MaxC > MinC)
	    factC = (double)(CO_ONE-1)/(MaxC - MinC);
          else
	    factC = 1.0;
	  co2 = CO_ONE >> 1;
	 }
       else
	 dop = doubData;
       for(i=0; i< num; i++)
          {
           /* init element buffer */
           elem = ip;
           ip += offs;
           dp = elem1;
           numElem = 1;
           copyAll = False;
           /* One Element taken from ip */
           if((copyAll= CO_IntersectX(elem, &dp, MaxX, &copyAll,
              &numElem, True)) < 0)
             {
              if(numElem == 0) continue;
              /* switch to next buffer */
              elem = elem1;
              dp = elem2;
              elem1 = elem2;
              elem2 = elem;
              copyAll = False;
             }
           /* Maximum 2 Elements */
           actElem = numElem;
           ep = elem;
           for(j = 0; j < actElem; j++)
              {
               if((copyAll= CO_IntersectX(ep, &dp, MinX, &copyAll,
        	  &numElem, False)) < 0)
        	 {
        	  if(copyAll == CO_REMOVED && j == actElem-1)
        	    copyAll = False;
        	  else
        	    {
        	     copyAll = True;
        	     if(j)
        	       {
        	        delta = j*offs;
        	        memcpy(dp, elem2, delta*sizeof(double));
        	        dp += delta;
        	       } 
        	    }
        	 }
               ep += offs;
              }
           if(numElem == 0) continue;
           if(copyAll)
             {
              /* switch to next buffer */
              elem = elem1;
              dp = elem2;
              elem1 = elem2;
              elem2 = elem;
              copyAll = False;
             }
           /* Maximum 4 Elements */
           actElem = numElem;
           ep = elem;
           for(j = 0; j < actElem; j++)
              {
               if((copyAll= CO_IntersectY(ep, &dp, MaxY, &copyAll,
        	  &numElem, True)) < 0)
        	 {
        	  if(copyAll == CO_REMOVED && j == actElem-1)
        	    copyAll = False;
        	  else
        	    {
        	     copyAll = True;
        	     if(j)
        	       {
        	        delta = j*offs;
        	        memcpy(dp, elem2, delta*sizeof(double));
        	        dp += delta;
        	       } 
        	    }
        	 }
               ep += offs;
              }
           if(numElem == 0) continue;
           if(copyAll)
             {
              /* switch to next buffer */
              elem = elem1;
              elem1 = elem2;
              elem2 = elem;
              copyAll = False;
              dp = elem1;
             }
           /* Maximum 8 Elements */
           actElem = numElem;
           ep = elem;
           for(j = 0; j < actElem; j++)
              {
               if((copyAll= CO_IntersectY(ep, &dp, MinY, &copyAll,
        	  &numElem, False)) < 0)
        	 {
        	  if(copyAll == CO_REMOVED && j == actElem-1)
        	    copyAll = False;
        	  else
        	    {
        	     copyAll = True;
        	     if(j)
        	       {
        	        delta = j*offs;
        	        memcpy(dp, elem2, delta*sizeof(double));
        	        dp += delta;
        	       } 
        	    }
        	 }
               ep += offs;
              }
           if(numElem == 0) continue;
           if(copyAll)
             {
              /* switch to next buffer */
              elem = elem1;
              elem1 = elem2;
              elem2 = elem;
              copyAll = False;
              dp = elem1;
             }
           /* Maximum 16 Elements */
           actElem = numElem;
           ep = elem;
           for(j = 0; j < actElem; j++)
              {
               if((copyAll= CO_IntersectZ(ep, &dp, MaxZ, &copyAll,
        	  &numElem, True)) < 0)
        	 {
        	  if(copyAll == CO_REMOVED && j == actElem-1)
        	    copyAll = False;
        	  else
        	    {
        	     copyAll = True;
        	     if(j)
        	       {
        	        delta = j*offs;
        	        memcpy(dp, elem2, delta*sizeof(double));
        	        dp += delta;
        	       } 
        	    }
        	 }
               ep += offs;
              }
           if(numElem == 0) continue;
           if(copyAll)
             {
              /* switch to next buffer */
              elem = elem1;
              elem1 = elem2;
              elem2 = elem;
              copyAll = False;
              dp = elem1;
             }
           /* Maximum 32 Elements */
           actElem = numElem;
           ep = elem;
           for(j = 0; j < actElem; j++)
              {
               if((copyAll= CO_IntersectZ(ep, &dp, MinZ, &copyAll,
        	  &numElem, False)) < 0)
        	 {
        	  if(copyAll == CO_REMOVED && j == actElem-1)
        	    copyAll = False;
        	  else
        	    {
        	     copyAll = True;
        	     if(j)
        	       {
        	        delta = j*offs;
        	        memcpy(dp, elem2, delta*sizeof(double));
        	        dp += delta;
        	       } 
        	    }
        	 }
               ep += offs;
              }
           if(numElem == 0) continue;
           if(copyAll)
              elem = elem1;
           /*
            * Now insert the elements
            */
           ep = elem;
	   for(j=0; j< numElem; j++)
	      {
	       actSize--;
	       if(actSize < 0)
	         {
	          actSize = 128;
		  Size += 128*offs;
		  if(type)
                    {
	             op = realloc(data, Size*sizeof(myCoor));
	             if(op == NULL)
	               {
	                free(data);
	                free(elem1);
	                free(elem2);
	                DDI.data = NULL;
			return(&DDI);
		       }
		     data = op;
		     op = data+oNorm*offs;
		    }
                  else
		    {
	             dop = realloc(doubData, Size*sizeof(double));
	             if(dop == NULL)
	               {
	                free(doubData);
	                free(elem1);
	                free(elem2);
	                DDI.doubData = NULL;
			return(&DDI);
		       }
		     doubData = dop;
		     dop = doubData+oNorm*offs;
		    }
		 }
	       oNorm++;
	       if(type)
		 {
		  *op++ = (myCoor)(factX*(*ep++ - MinX))-co2;
		  *op++ = (myCoor)(factY*(*ep++ - MinY))-co2;
		  *op++ = (myCoor)(factZ*(*ep++ - MinZ))-co2;
		  *op++ = (myCoor)(factX*(*ep++ - MinX))-co2;
		  *op++ = (myCoor)(factY*(*ep++ - MinY))-co2;
		  *op++ = (myCoor)(factZ*(*ep++ - MinZ))-co2;
		  *op++ = (myCoor)(factX*(*ep++ - MinX))-co2;
		  *op++ = (myCoor)(factY*(*ep++ - MinY))-co2;
		  *op++ = (myCoor)(factZ*(*ep++ - MinZ))-co2;
		  /*
	           * Color Info
	           */
		  for(k=0; k < COLOR_3D; k++)
	             *op++ = (myCoor)(factC*(*ep++ - MinC));
		 }
	       else
		 {
		  /* in case of double just copy */
                  for(k=0; k<offs; k++)
		     *dop++ = *ep++;
		 }
	      }
          }
      /*
       * reallocate memory
       */
       Size = oNorm*offs;
       DDI.norm = oNorm;
       if(type)
         {
          op = realloc(data, Size*sizeof(myCoor));
          if(op == NULL)
	    {
	     free(data);
	     free(elem1);
	     free(elem2);
	     DDI.data = NULL;
	     return(&DDI);
	    }
          data = op;
          DDI.FactX = 1.0/factX;
          DDI.FactY = 1.0/factY;
          DDI.FactZ = 1.0/factZ;
	  DDI.FactC = 1.0/factC;
	 }
       else
	 {
          dop = realloc(doubData, Size*sizeof(double));
          if(dop == NULL)
	    {
	     free(doubData);
	     free(elem1);
	     free(elem2);
	     DDI.doubData = NULL;
	     return(&DDI);
	    }
          doubData = dop;
	 }
      }
    DDI.data = data;
    DDI.doubData = doubData;
    free(elem1);
    free(elem2);
    return(&DDI);
   }
/*
 * format from double to integer, values
 * outside XDim->Max, XDim->Min are cut off
 */
   
int CO_Data2Int(ViewportInfo *view, XDimInfo *XDim)
   {
    int i, Size;
    myCoor *ip;
    double *dp, fact, Max, Min, val;
    unsigned char *cp;
    my3DInfo *DDI;

    if(view->actData || view->charData)
      {
       if(view->Data)
	 free(view->Data);
       if(view->dataType < MY_TRIANGLE)
        {
         Size = view->Width * view->Height;
         view->Data = malloc(Size*sizeof(myCoor));
	 ip = view->Data;
	 Min = view->Min;
	 Max = view->Max;
	 if(Max > Min)
	   fact = (double)(CO_ONE-1)/(Max - Min);
	 else
	   fact = 1.0;
	 view->FactZ = 1.0/fact;
	 if(!view->Data)
	   {
	    sprintf(XDim->error, XDERR_NO_MEM_D,
	       Size*(int)sizeof(myCoor));
	    return(ERROR);
	   }
	  switch (view->dataType)
           {
            case MY_UCHAR:
	    cp = view->charData;
	    view->dMin = 0;
	    for(i=0; i<Size; i++)
	       {
		val = (double)*cp++;
		val = Mminmax(val,Min,Max);
		*ip++ = (myCoor)(fact * (val - Min));
	       }
            break;
            default:
	    dp = view->actData;
	    if(Min > 0.0 || Max < 0.0)
	      {
	       view->dMin = 0;
	       for(i=0; i<Size; i++)
		  {
		   val = *dp++;
		   val = Mminmax(val,Min,Max);
		   *ip++ = (myCoor)(fact * (val - Min));
		  }
	      }
	    else
	      {
	       view->dMin = fact*Min;
	       for(i=0; i<Size; i++)
		  {
		   val = *dp++;
		   val = Mminmax(val,Min,Max);
		   *ip++ = (myCoor)(fact * val);
		  }
	      }
	    break;
	   }
        }
      else
        {
	 DDI = CO_Triangle2Int(view->actData, view->Width, view->Height,
		MY_INT, view->OffX, view->EndX, view->OffY, view->EndY,
		view->Min, view->Max, view->MinC, view->MaxC);
	 view->dMin = -(CO_ONE >> 1);
	 if(DDI->norm < 0)
	   {
	    strcpy(XDim->error, XDERR_NO_MEM);
	    return(ERROR);
	   }
	 else
	   {
	    view->Data = DDI->data;
	    view->nInter = DDI->inter;
	    view->nNorm = DDI->norm;
	    view->nLine = DDI->lines;
	    view->FactX = DDI->FactX;
	    view->FactY = DDI->FactY;
	    view->FactZ = DDI->FactZ;
	    view->FactC = DDI->FactC;
	   }
        }
      }
    view->changed &= ~DATA_CHANGED;
    return(OK);
   }
/*
 * Return the Value depending on the Data Type
 */

double CO_GetVal(int x, int y, ViewportInfo *view)
   {
    double val;
    
    switch (view->dataType)
       {
        case MY_UCHAR:
        val = (double) *(view->charData+ view->Width*y+ x);
        break;
        default:
        val = *(view->actData+ view->Width*y+ x);
	break;
       }
    return(val);
   }

/*
 * do an interpolation
 */

double CO_Inter(double xs, double ys, ViewportInfo *view)
   {
    int del, dx, dy;
    double min,max,*dp, z1, z2, zn, fact;

    del = view->Width;
    min = view->Min;
    max = view->Max;
    dx = (int) xs;
    dy = (int) ys;
    if(view->dataType)
      zn = CO_GetVal(dx, dy, view);
    else
      {
       dp = view->actData + dy*del+ dx;
       zn = *dp;
      }
    zn = Mminmax(zn,min,max);
    fact = xs-floor(xs);
    if(dx<del-1)
      {
       if(view->dataType)
	 z1 = CO_GetVal(dx+1, dy, view);
       else
	 z1 = *(dp+1);
       z1 = Mminmax(z1,min,max);
       z1 = zn + fact*(z1-zn);
      }
    else
      z1 = zn;
    if(dy < view->Height-1)
      {
       if(view->dataType)
         {
          dy++;
	  zn = CO_GetVal(dx, dy, view);
	 }
       else
         {
	  dp += del;
	  zn = *dp;
         }
       zn = Mminmax(zn,min,max);
      }
    if(dx<del-1)
      {
       if(view->dataType)
	 z2 = CO_GetVal(dx+1, dy, view);
       else
         z2 = *(dp+1);
       z2 = Mminmax(z2,min,max);
       z2 = zn + fact*(z2-zn);
      }    
    else
      z2 = zn;
    fact = ys -floor(ys);
    zn = z1 + fact*(z2-z1);
    return(zn);
   }

/*
 * Start spline interpolation
 */

int CO_StartSpline(int len, double dx, double *a, double *z, XDimInfo *XDim)
   {
    int ret, i, len1, size;
    double fact, dx2, dx23, dx6,*b,*c,*d;
    
    size = len*sizeof(double);
    b = (double *)malloc(size);
    c = (double *)malloc(size);
    d = (double *)malloc(size);
    if(len < 3)
      {
       sprintf(XDim->error, XDERR_SPLINE_DATA);
       free(a);
       ret = ERROR;
      }
    else if(b== NULL || c == NULL || d == NULL)
      {
       sprintf(XDim->error, XDERR_NO_MEM_D, 3*size);
       free(a);
       ret = ERROR;
      }
    else
      {
       dx2 = dx*dx;
       len1 = len-1;
       a[0] = 0.0;
       b[0] = 1.0;
       c[0] = 0.0;
       a[len1] = 0.0;
       b[len1] = 1.0;
       c[len1] = 0.0;
       d[0] = (z[1]-z[0])/dx2;
       d[len1] = (z[len1-1]-z[len1])/dx2;
       dx6 = dx2/6.0;
       dx23 = 2.0*dx2/3.0;
       for(i=1; i<len1; i++)
          {
           a[i] = dx6;
           b[i] = dx23;
           c[i] = dx6;
           d[i] = z[i+1]-2*z[i]+z[i-1];
          }
       for(i=1; i<len1; i++)
          {
           fact = -a[i]/b[i-1];
           b[i] += c[i-1]*fact;
           d[i] += d[i-1]*fact;
          }
       a[len1] = d[len1]/b[len1];
       a[0] = d[0]/b[0];
       for(i= len1-1; i>0; i--)
          {
           a[i] = (d[i]-c[i]*a[i+1])/b[i];
          } 
       ret = OK;
      }
    free(b);
    free(c);
    free(d);
    return(ret);
   }

/*
 * do spline interpolation
 */

double CO_Spline(double x, double dx, int len, double *xvec, double *yvec)
   {
    double y,xi,xi1, dx6;
    int frac;
    
    frac = ((int)x);
    xi1 = x- (double) frac;
    frac++;
    xi = x-(double) frac;
    dx6 = 1.0/dx/6.0;
    if(x < 0.0)
      {
       y = yvec[0];
      }
    else if(frac>=len)
      {
       y = yvec[len-1];
      }
    else
      {
       y = xvec[frac]*xi1*(xi1*xi1*dx6-dx/6.0)+yvec[frac]*xi1/dx-
           xvec[frac-1]*xi*(xi*xi*dx6-dx/6.0)-yvec[frac-1]*xi/dx;
      }
    return(y);
   }

/*
 * Cut a line through a grid
 */

int CO_NextStep(double x1, double y1, double *x, double *y, double *z,
	ViewportInfo *view)
   {
    int Next;
    double xs, ys, xn, yn, xn1, yn1;

    xn = *x;
    yn = *y;
    xn1 = floor(xn);
    yn1 = floor(yn);
    if(x1 >= xn1 && x1 < xn1+1.0 && y1 >= yn1 && y1 < yn1+1.0)
      Next = False;
    else
      Next = True;  
    xs = x1;
    ys = y1;
    if(xn1>=x1)
      {
       if(xn1>=xn)
         xn1 -= 1.0;
       if(xn1 >= x1)
         xs = xn1;
      }
    else
      {
       xn1 += 1.0;
       if(xn1<=x1)
          xs = xn1;
      }
    if(xn != x1)
      ys = y1 + (yn - y1)*(xs-x1)/(xn-x1);
    if(ys < yn1 || ys > yn1+1.0)
      {
       if(yn1>=y1)
         {
          if(yn1>=yn)
            yn1 -= 1.0;
          if(yn1>=y1)
	    ys = yn1;
	 }
       else
	 {
	  yn1 += 1.0;
	  if(yn1<=y1)
             ys = yn1;
	 }
       if(yn != y1)
	 xs = x1 + (xn - x1)*(ys-y1)/(yn-y1);
      }
    /* Now do Interpolation */
    
    *z = CO_Inter(xs, ys, view);
    *y = ys;
    *x = xs;
    return(Next);
   }

/*
 * Calculate Coordinates from screen x,y
 */

int CO_Point(int x, int y, double *xr, double *yr, double *zr,
	ViewportInfo *view, XDimInfo *XDim)
   {
    int next, found;
    double fact, Sub, cx, cy, cz;
    myCoor x0, y0, x1, y1, coMaxX, sinXY, cosXY, sinZ, cosZ, moveX, moveY;
    myCoor z1, MaxX, MaxY, coMaxY, zoomZ, zoomX, zoomY, height, width;
    myCoor x2, y2, x3, y3, dumx, dumy, Fact;
    double pi2, dAngle, cx1, cy1, cz1, cx2, cy2, Max, Min;
    int xa, ya, xas, yas, xb, yb, xbs, ybs, xc, yc, xcs, ycs, xd, yd, xds, yds;
    
    dAngle = 2.0*PI;
    pi2 = PI/2.0;
    sinXY = view->sinXY;
    cosXY = view->cosXY;
    sinZ = view->sinZ;
    cosZ = view->cosZ;
    moveX = view->moveX;
    moveY = view->moveY;
    zoomX = view->zoomX;
    zoomY = view->zoomY;
    zoomZ = view->zoomZ;
    MaxX = view->Width-1;
    MaxY = view->Height-1;
    width = view->pwidth;
    height = view->pheight;
    y = Mupdown(y,height);
    if(MaxY > MaxX)
      {
       if(MaxY >0)
	 Fact = MaxY;
       else
	 Fact = 1;
       coMaxY = CO_ONE >> 1;
       coMaxX = (MaxX<<(CO_ONESHIFT-1))/Fact;
      }
    else
      {
       if(MaxX >0)
	 Fact = MaxX;
       else
	 Fact = 1;
       coMaxX = CO_ONE >> 1;
       coMaxY = (MaxY<<(CO_ONESHIFT-1))/Fact;
      }


    if((dAngle -= pi2) <= view->rotXY)
      {
       x0 = -coMaxX;
       y0 = -coMaxY;
       xas = 0;
       yas = 0;
       x1 = coMaxX;
       y1 = -coMaxY;
       xbs = MaxX;
       ybs = 0;
       x2 = coMaxX;
       y2 = coMaxY;
       xcs = MaxX;
       ycs = MaxY;
       x3 = -coMaxX;
       y3 = coMaxY;
       xds = 0;
       yds = MaxY;
      }
    else if((dAngle -= pi2) <= view->rotXY)
      {
       x0 = -coMaxX;
       y0 = coMaxY;
       xas = 0;
       yas = MaxY;
       x1 = -coMaxX;
       y1 = -coMaxY;
       xbs = 0;
       ybs = 0;
       x2 = coMaxX;
       y2 = -coMaxY;
       xcs = MaxX;
       ycs = 0;
       x3 = coMaxX;
       y3 = coMaxY;
       xds = MaxX;
       yds = MaxY;
      }
    else if((dAngle -= pi2) <= view->rotXY)
      {
       x0 = coMaxX;
       y0 = coMaxY;
       xas = MaxX;
       yas = MaxY;
       x1 = -coMaxX;
       y1 = coMaxY;
       xbs = 0;
       ybs = MaxY;
       x2 = -coMaxX;
       y2 = -coMaxY;
       xcs = 0;
       ycs = 0;
       x3 = coMaxX;
       y3 = -coMaxY;
       xds = MaxX;
       yds = 0;
      }
    else
      {
       x0 = coMaxX;
       y0 = -coMaxY;
       xas = MaxX;
       yas = 0;
       x1 = coMaxX;
       y1 = coMaxY;
       xbs = MaxX;
       ybs = MaxY;
       x2 = -coMaxX;
       y2 = coMaxY;
       xcs = 0;
       ycs = MaxY;
       x3 = -coMaxX;
       y3 = -coMaxY;
       xds = 0;
       yds = 0;
      }
    CO_TransformXY(xa,ya,dumx,dumy,x0,y0,0,sinXY,cosXY,sinZ,cosZ,
	zoomX,zoomY,zoomZ,moveX,moveY,0);
    CO_TransformXY(xb,yb,dumx,dumy,x1,y1,0,sinXY,cosXY,sinZ,cosZ,
	zoomX,zoomY,zoomZ,moveX,moveY,0);
    CO_TransformXY(xc,yc,dumx,dumy,x2,y2,0,sinXY,cosXY,sinZ,cosZ,
	zoomX,zoomY,zoomZ,moveX,moveY,0);
    CO_TransformXY(xd,yd,dumx,dumy,x3,y3,0,sinXY,cosXY,sinZ,cosZ,
	zoomX,zoomY,zoomZ,moveX,moveY,0);
    if(x<xd || x> xb)
       return(False);;
    if(x < xa)
      {
       if(xa != xd)
         {
          cx1 = xds + ((double)((x-xd)*(xas-xds)))/(xa-xd);
          cy1 = yds + ((double)((x-xd)*(yas-yds)))/(xa-xd);
         }
       else
         {
          cx1 = xds;
          cy1 = yds;
         }
      }
    else
      {
       if(xb != xa)
         {
          cx1 = xas + ((double)((x-xa)*(xbs-xas)))/(xb-xa);
          cy1 = yas + ((double)((x-xa)*(ybs-yas)))/(xb-xa);
         }
       else
         {
          cx1 = xas;
          cy1 = yas;
         }
      }
    if(x < xc)
      {
       if(xc != xd)
         {
          cx2 = xds + ((double)((x-xd)*(xcs-xds)))/(xc-xd);
          cy2 = yds + ((double)((x-xd)*(ycs-yds)))/(xc-xd);
         }
       else
         {
          cx2 = xds;
          cy2 = yds;
         }
      }
    else
      {
       if(xb != xc)
         {
          cx2 = xcs + ((double)((x-xc)*(xbs-xcs)))/(xb-xc);
          cy2 = ycs + ((double)((x-xc)*(ybs-ycs)))/(xb-xc);
         }
       else
         {
          cx2 = xcs;
          cy2 = ycs;
         }
      }
    cz1 = CO_Inter(cx1, cy1, view);
    cx = cx1;
    cy = cy1;
    cz = cz1;
    Min = view->Min;
    Max = view->Max;
    if(Max > Min)
      fact = (CO_ONE-1)/(Max - Min);
    else
      fact = 1.0;
    x1 = (myCoor)((cx/Fact)*CO_ONE)-coMaxX;
    y1 = (myCoor)((cy/Fact)*CO_ONE)-coMaxY;
    if(Min <0.0 && Max > 0.0)
      Sub = 0.0;
    else
      Sub = Min;
    z1 = (myCoor)(fact * (cz - Sub));
    CO_TransformXY(xb,yb,dumx,dumy,x1,y1,z1,sinXY,cosXY,sinZ,cosZ,
	zoomX,zoomY,zoomZ,moveX,moveY,0);
    found = False;
    do
      {
       cx1 = cx;
       cy1 = cy;
       cz1 = cz;
       ya = yb;
       next=CO_NextStep(cx2,cy2,&cx,&cy,&cz, view);
       x1 = (myCoor)((cx/Fact)*CO_ONE)-coMaxX;
       y1 = (myCoor)((cy/Fact)*CO_ONE)-coMaxY;
       z1 = (myCoor)(fact * (cz - Sub));
       CO_TransformXY(xb,yb,dumx,dumy,x1,y1,z1,sinXY,cosXY,sinZ,cosZ,
	zoomX,zoomY,zoomZ,moveX,moveY,0);
       if((y <= yb && y >= ya) || (y <= ya && y >= yb))
         {
          found = True;
          if(ya!=yb)
            {
	     *xr = cx1 + ((cx-cx1)*(y-ya))/(yb-ya);
	     *yr = cy1 + ((cy-cy1)*(y-ya))/(yb-ya);
	     *zr = cz1 + ((cz-cz1)*(y-ya))/(yb-ya);
	    }
	  else
	    {
	     *xr = cx1;
	     *yr = cy1;
	     *zr = cz1;
	    }            
          break;
         }
      }
    while(next);
    return(found);
   }

/*
 * Calculate Axis Base values
 */

void CO_CalculateBases(ViewportInfo *view2)
   {
    switch (view2->dataType)
      {
       case MY_TRIANGLE:
       CO_MinMaxTri(&(view2->EndX), &(view2->OffX), 0, view2->Height,
		view2->Width, view2->actData);
       CO_MinMaxTri(&(view2->EndY), &(view2->OffY), 1, view2->Height,
		view2->Width, view2->actData);
       CO_MinMaxTri(&(view2->Max), &(view2->Min), 2, view2->Height,
		view2->Width, view2->actData);
       CO_MinMaxCol3D(&(view2->MaxC), &(view2->MinC), view2->Height,
		view2->Width, view2->actData);
       break;
       case MY_UCHAR:
       CO_MinMaxChar(&(view2->Max), &(view2->Min),
	  view2->Width * view2->Height,view2->charData);
       break;
       default:
       CO_MinMax(&(view2->Max), &(view2->Min),
	  view2->Width * view2->Height,view2->actData);
       break;
      }
    if(view2->dataType < MY_TRIANGLE)
      {
       view2->baseX = CO_Base(view2->Width*view2->FactX
	   +view2->OffX);
       view2->baseY = CO_Base(view2->Height*view2->FactY
	   +view2->OffY);
      }
    else
      {
       if(fabs(view2->OffX) > fabs(view2->EndX))
	 view2->baseX = CO_Base(view2->OffX);
       else
	 view2->baseX = CO_Base(view2->EndX);
       if(fabs(view2->OffY) > fabs(view2->EndY))
	 view2->baseY = CO_Base(view2->OffY);
       else
	 view2->baseY = CO_Base(view2->EndY);
      }
    if(fabs(view2->Max) > fabs(view2->Min))
      view2->baseZ = CO_Base(view2->Max);
    else
      view2->baseZ = CO_Base(view2->Min);
   }
